@@ -36,20 +36,28 @@
 #include <net/if.h>
 #include <unistd.h>
 #include <errno.h>
+#include <sys/time.h>
 
 #include "private.h"
+#include "useful.h"
 #include "timeout.h"
 #include "time.h"
 #include "io.h"
 #include "time-private.h"
 #include "queue.h"
 #include "net.h"
+#include "net-private.h"
 #include "netlink.h"
 #include "rtnl.h"
 #include "missing.h"
 #include "icmp6.h"
 #include "icmp6-private.h"
 
+/* RFC4191 */
+#ifndef ND_OPT_ROUTE_INFORMATION
+#define ND_OPT_ROUTE_INFORMATION	24
+#endif
+
 #define CLIENT_DEBUG(fmt, args...)					\
 	l_util_debug(client->debug_handler, client->debug_data,		\
 			"%s:%i " fmt, __func__, __LINE__, ## args)
@@ -125,6 +133,10 @@ static int icmp6_open_router_common(const struct icmp6_filter *filter,
 				ifr.ifr_name, strlen(ifr.ifr_name) + 1);
 	}
 
+	if (r < 0)
+		goto fail;
+
+	r = setsockopt(s, SOL_SOCKET, SO_TIMESTAMP, &yes, sizeof(yes));
 	if (r < 0)
 		goto fail;
 
@@ -190,9 +202,11 @@ static int icmp6_send_router_solicitation(int s, const uint8_t mac[static 6])
 	return 0;
 }
 
-static int icmp6_receive(int s, void *buf, size_t buf_len, struct in6_addr *src)
+static int icmp6_receive(int s, void *buf, size_t buf_len, struct in6_addr *src,
+				uint64_t *out_timestamp)
 {
-	char c_msg_buf[CMSG_SPACE(sizeof(int))];
+	char c_msg_buf[CMSG_SPACE(sizeof(int)) +
+			CMSG_SPACE(sizeof(struct timeval))];
 	struct iovec iov = {
 		.iov_base = buf,
 		.iov_len = buf_len,
@@ -209,6 +223,7 @@ static int icmp6_receive(int s, void *buf, size_t buf_len, struct in6_addr *src)
 	};
 	struct cmsghdr *cmsg;
 	ssize_t l;
+	uint64_t timestamp = 0;
 
 	l = recvmsg(s, &msg, MSG_DONTWAIT);
 	if (l < 0)
@@ -229,13 +244,27 @@ static int icmp6_receive(int s, void *buf, size_t buf_len, struct in6_addr *src)
 
 			if (hops != 255)
 				return -EMULTIHOP;
+		} else if (cmsg->cmsg_level == SOL_SOCKET &&
+				cmsg->cmsg_type == SCM_TIMESTAMP &&
+				cmsg->cmsg_len ==
+				CMSG_LEN(sizeof(struct timeval))) {
+			const struct timeval *tv = (void *) CMSG_DATA(cmsg);
+
+			timestamp = _time_realtime_to_boottime(tv);
 		}
 	}
 
 	memcpy(src, &saddr.sin6_addr, sizeof(saddr.sin6_addr));
+	*out_timestamp = timestamp ?: l_time_now();
 	return 0;
 }
 
+struct icmp6_event_handler_entry {
+	l_icmp6_client_event_cb_t handle;
+	void *user_data;
+	l_icmp6_destroy_cb_t destroy;
+};
+
 struct l_icmp6_client {
 	uint32_t ifindex;
 	uint8_t mac[6];
@@ -248,9 +277,7 @@ struct l_icmp6_client {
 	uint32_t route_priority;
 	struct l_queue *routes;
 
-	l_icmp6_client_event_cb_t event_handler;
-	void *event_data;
-	l_icmp6_destroy_cb_t event_destroy;
+	struct l_queue *event_handlers;
 
 	l_icmp6_debug_cb_t debug_handler;
 	l_icmp6_destroy_cb_t debug_destroy;
@@ -261,10 +288,17 @@ struct l_icmp6_client {
 };
 
 static inline void icmp6_client_event_notify(struct l_icmp6_client *client,
-						enum l_icmp6_client_event event)
+					enum l_icmp6_client_event event,
+					void *event_data)
 {
-	if (client->event_handler)
-		client->event_handler(client, event, client->event_data);
+	const struct l_queue_entry *entry;
+
+	for (entry = l_queue_get_entries(client->event_handlers); entry;
+			entry = entry->next) {
+		struct icmp6_event_handler_entry *handler = entry->data;
+
+		handler->handle(client, event, event_data, handler->user_data);
+	}
 }
 
 static bool icmp6_client_remove_route(void *data, void *user_data)
@@ -274,7 +308,7 @@ static bool icmp6_client_remove_route(void *data, void *user_data)
 
 	if (client->rtnl)
 		l_rtnl_route_delete(client->rtnl, client->ifindex, r,
-				NULL, NULL, NULL);
+					NULL, NULL, NULL);
 
 	l_free(r);
 	return true;
@@ -304,16 +338,24 @@ static void icmp6_client_setup_routes(struct l_icmp6_client *client)
 		l_rtnl_route_add(client->rtnl, client->ifindex, rt,
 					NULL, NULL, NULL);
 
-	for (i = 0; i < ra->n_prefixes; i++) {
-		struct route_info *info = &ra->prefixes[i];
+	for (i = 0; i < ra->n_routes; i++) {
+		char prefix_buf[INET6_ADDRSTRLEN];
+		struct route_info *info = &ra->routes[i];
 
 		if (info->valid_lifetime == 0)
 			continue;
 
-		if (!inet_ntop(AF_INET6, info->address, buf, sizeof(buf)))
+		if (!inet_ntop(AF_INET6, info->address, prefix_buf,
+				sizeof(prefix_buf)))
 			continue;
 
-		rt = l_rtnl_route_new_prefix(buf, info->prefix_len);
+		if (info->onlink)
+			rt = l_rtnl_route_new_prefix(prefix_buf,
+							info->prefix_len);
+		else
+			rt = l_rtnl_route_new_static(buf, prefix_buf,
+							info->prefix_len);
+
 		if (!rt)
 			continue;
 
@@ -332,17 +374,25 @@ static void icmp6_client_setup_routes(struct l_icmp6_client *client)
 static int icmp6_client_handle_message(struct l_icmp6_client *client,
 						struct nd_router_advert *ra,
 						size_t len,
-						const struct in6_addr *src)
+						const struct in6_addr *src,
+						uint64_t timestamp)
 {
-	struct l_icmp6_router *r = _icmp6_router_parse(ra, len, src->s6_addr);
+	struct l_icmp6_router *r =
+		_icmp6_router_parse(ra, len, src->s6_addr, timestamp);
 
 	if (!r)
 		return -EBADMSG;
 
+	icmp6_client_event_notify(client,
+					L_ICMP6_CLIENT_EVENT_ROUTER_FOUND,
+					r);
+
+	/* DHCP6 client may have stopped us */
+	if (!client->io)
+		return -ECANCELED;
+
 	if (!client->ra) {
 		client->ra = r;
-		icmp6_client_event_notify(client,
-					L_ICMP6_CLIENT_EVENT_ROUTER_FOUND);
 		icmp6_client_setup_routes(client);
 		return 0;
 	}
@@ -351,8 +401,7 @@ static int icmp6_client_handle_message(struct l_icmp6_client *client,
 	 * TODO: Figure out if the RA has updated info and update routes
 	 * accordingly.
 	 */
-	_icmp6_router_free(client->ra);
-	client->ra = r;
+	_icmp6_router_free(r);
 	return 0;
 }
 
@@ -363,6 +412,8 @@ static bool icmp6_client_read_handler(struct l_io *io, void *userdata)
 	struct nd_router_advert *ra;
 	ssize_t l;
 	struct in6_addr src;
+	int r;
+	uint64_t timestamp = 0;
 
 	/* Poke to see how many bytes we need to read / alloc */
 	l = recv(s, NULL, 0, MSG_PEEK|MSG_TRUNC);
@@ -373,7 +424,7 @@ static bool icmp6_client_read_handler(struct l_io *io, void *userdata)
 	}
 
 	ra = l_malloc(l);
-	if (icmp6_receive(s, ra, l, &src) < 0)
+	if (icmp6_receive(s, ra, l, &src, &timestamp) < 0)
 		goto done;
 
 	if ((size_t) l < sizeof(struct nd_router_advert)) {
@@ -381,7 +432,8 @@ static bool icmp6_client_read_handler(struct l_io *io, void *userdata)
 		goto done;
 	}
 
-	if (icmp6_client_handle_message(client, ra, l, &src) < 0)
+	r = icmp6_client_handle_message(client, ra, l, &src, timestamp);
+	if (r < 0)
 		goto done;
 
 	/* Stop solicitations */
@@ -395,9 +447,10 @@ done:
 }
 
 static void icmp6_client_timeout_send(struct l_timeout *timeout,
-								void *user_data)
+							void *user_data)
 {
-	static const uint64_t MAX_SOLICITATION_INTERVAL = 3600 * L_MSEC_PER_SEC;
+	static const uint64_t MAX_SOLICITATION_INTERVAL =
+							3600 * L_MSEC_PER_SEC;
 	static const uint64_t SOLICITATION_INTERVAL = 4 * L_MSEC_PER_SEC;
 	struct l_icmp6_client *client = user_data;
 	int r;
@@ -436,6 +489,16 @@ LIB_EXPORT struct l_icmp6_client *l_icmp6_client_new(uint32_t ifindex)
 	return client;
 }
 
+static void icmp6_event_handler_destroy(void *data)
+{
+	struct icmp6_event_handler_entry *handler = data;
+
+	if (handler->destroy)
+		handler->destroy(handler->user_data);
+
+	l_free(handler);
+}
+
 LIB_EXPORT void l_icmp6_client_free(struct l_icmp6_client *client)
 {
 	if (unlikely(!client))
@@ -443,6 +506,8 @@ LIB_EXPORT void l_icmp6_client_free(struct l_icmp6_client *client)
 
 	l_icmp6_client_stop(client);
 	l_queue_destroy(client->routes, NULL);
+	l_icmp6_client_set_debug(client, NULL, NULL, NULL);
+	l_queue_destroy(client->event_handlers, icmp6_event_handler_destroy);
 	l_free(client);
 }
 
@@ -464,15 +529,16 @@ LIB_EXPORT bool l_icmp6_client_start(struct l_icmp6_client *client)
 		return false;
 
 	if (!client->have_mac) {
-		if (!l_net_get_mac_address(client->ifindex, client->mac)) {
-			close(s);
-			return false;
-		}
+		if (!l_net_get_mac_address(client->ifindex, client->mac))
+			goto err;
 
 		client->have_mac = true;
 	}
 
 	client->io = l_io_new(s);
+	if (!client->io)
+		goto err;
+
 	l_io_set_close_on_destroy(client->io, true);
 	l_io_set_read_handler(client->io, icmp6_client_read_handler,
 					client, NULL);
@@ -488,6 +554,10 @@ LIB_EXPORT bool l_icmp6_client_start(struct l_icmp6_client *client)
 		icmp6_client_timeout_send(client->timeout_send, client);
 
 	return true;
+
+err:
+	close(s);
+	return false;
 }
 
 LIB_EXPORT bool l_icmp6_client_stop(struct l_icmp6_client *client)
@@ -527,20 +597,24 @@ LIB_EXPORT const struct l_icmp6_router *l_icmp6_client_get_router(
 	return client->ra;
 }
 
-LIB_EXPORT bool l_icmp6_client_set_event_handler(struct l_icmp6_client *client,
+LIB_EXPORT bool l_icmp6_client_add_event_handler(struct l_icmp6_client *client,
 					l_icmp6_client_event_cb_t handler,
-					void *userdata,
+					void *user_data,
 					l_icmp6_destroy_cb_t destroy)
 {
+	struct icmp6_event_handler_entry *handler_entry;
+
 	if (unlikely(!client))
 		return false;
 
-	if (client->event_destroy)
-		client->event_destroy(client->event_data);
+	if (!client->event_handlers)
+		client->event_handlers = l_queue_new();
 
-	client->event_handler = handler;
-	client->event_data = userdata;
-	client->event_destroy = destroy;
+	handler_entry = l_new(struct icmp6_event_handler_entry, 1);
+	handler_entry->handle = handler;
+	handler_entry->user_data = user_data;
+	handler_entry->destroy = destroy;
+	l_queue_push_head(client->event_handlers, handler_entry);
 
 	return true;
 }
@@ -598,8 +672,9 @@ LIB_EXPORT bool l_icmp6_client_set_rtnl(struct l_icmp6_client *client,
 	return true;
 }
 
-LIB_EXPORT bool l_icmp6_client_set_route_priority(struct l_icmp6_client *client,
-							uint32_t priority)
+LIB_EXPORT bool l_icmp6_client_set_route_priority(
+						struct l_icmp6_client *client,
+						uint32_t priority)
 {
 	if (unlikely(!client))
 		return false;
@@ -617,18 +692,19 @@ struct l_icmp6_router *_icmp6_router_new()
 
 void _icmp6_router_free(struct l_icmp6_router *r)
 {
-	l_free(r->prefixes);
+	l_free(r->routes);
 	l_free(r);
 }
 
 struct l_icmp6_router *_icmp6_router_parse(const struct nd_router_advert *ra,
 						size_t len,
-						const uint8_t src[static 16])
+						const uint8_t src[static 16],
+						uint64_t timestamp)
 {
 	struct l_icmp6_router *r;
 	const uint8_t *opts;
 	uint32_t opts_len;
-	uint32_t n_prefixes = 0;
+	uint32_t n_routes = 0;
 
 	if (ra->nd_ra_type != ND_ROUTER_ADVERT)
 		return NULL;
@@ -665,7 +741,35 @@ struct l_icmp6_router *_icmp6_router_parse(const struct nd_router_advert *ra,
 				return NULL;
 
 			if (opts[3] & ND_OPT_PI_FLAG_ONLINK)
-				n_prefixes += 1;
+				n_routes += 1;
+			break;
+		case ND_OPT_ROUTE_INFORMATION:
+			if (l < 8)
+				return NULL;
+
+			if (opts[2] > 128 || opts[2] > (l - 8) * 8)
+				return NULL;
+
+			/*
+			 * RFC 4191 Section 2.3:
+			 * "If the Reserved (10) value is received, the Route
+			 * Information Option MUST be ignored."
+			 */
+			if (bit_field(opts[3], 3, 2) == 2)
+				break;
+
+			/*
+			 * RFC 4191 Section 3.1:
+			 * "The Router Preference and Lifetime values in a ::/0
+			 * Route Information Option override the preference and
+			 * lifetime values in the Router Advertisement header."
+			 *
+			 * Don't count ::/0 routes.
+			 */
+			if (opts[2] == 0)
+				break;
+
+			n_routes += 1;
 			break;
 		}
 
@@ -673,11 +777,10 @@ struct l_icmp6_router *_icmp6_router_parse(const struct nd_router_advert *ra,
 		opts_len -= l;
 	}
 
-
 	r = _icmp6_router_new();
 	memcpy(r->address, src, sizeof(r->address));
-	r->prefixes = l_new(struct route_info, n_prefixes);
-	r->n_prefixes = n_prefixes;
+	r->routes = l_new(struct route_info, n_routes);
+	r->n_routes = n_routes;
 
 	if (ra->nd_ra_flags_reserved & ND_RA_FLAG_MANAGED)
 		r->managed = true;
@@ -689,11 +792,12 @@ struct l_icmp6_router *_icmp6_router_parse(const struct nd_router_advert *ra,
 	if (r->pref == 0x2) /* If invalid, reset to medium */
 		r->pref = 0;
 
+	r->start_time = timestamp;
 	r->lifetime = L_BE16_TO_CPU(ra->nd_ra_router_lifetime);
 
 	opts = (uint8_t *) (ra + 1);
 	opts_len = len - sizeof(struct nd_router_advert);
-	n_prefixes = 0;
+	n_routes = 0;
 
 	while (opts_len) {
 		uint8_t t = opts[0];
@@ -711,28 +815,121 @@ struct l_icmp6_router *_icmp6_router_parse(const struct nd_router_advert *ra,
 			break;
 		case ND_OPT_PREFIX_INFORMATION:
 		{
-			struct route_info *i = &r->prefixes[n_prefixes];
+			struct route_info *i = &r->routes[n_routes];
 
 			if (!(opts[3] & ND_OPT_PI_FLAG_ONLINK))
 				break;
 
 			i->prefix_len = opts[2];
+			i->onlink = true;
 			i->valid_lifetime = l_get_be32(opts + 4);
 			i->preferred_lifetime = l_get_be32(opts + 8);
-			memcpy(i->address, opts + 16, 16);
 
-			n_prefixes += 1;
+			/*
+			 * Only the initial Prefix Length bits of the prefix
+			 * are valid.  The remaining bits "MUST" be ignored
+			 * by the receiver.
+			 */
+			memcpy(i->address, net_prefix_from_ipv6(opts + 16,
+							i->prefix_len), 16);
+
+			/*
+			 * For SLAAC (RFC4862) we need to "silently ignore"
+			 * routes with a preferred lifetime longer than valid
+			 * lifetime, and those with the link-local prefix.
+			 * Since it makes sense, do it regardless of SLAAC.
+			 */
+			if (i->preferred_lifetime > i->valid_lifetime)
+				break;
+
+			if (i->prefix_len >= 10 &&
+					IN6_IS_ADDR_LINKLOCAL(i->address))
+				break;
+
+			n_routes += 1;
+			break;
+		}
+		case ND_OPT_RTR_ADV_INTERVAL:
+			if (l < 8)
+				break;
+
+			r->max_rtr_adv_interval_ms = l_get_be32(opts + 4);
+			break;
+		case ND_OPT_ROUTE_INFORMATION:
+		{
+			struct route_info *i = &r->routes[n_routes];
+			uint8_t preference = bit_field(opts[3], 3, 2);
+
+			if (preference == 2)
+				break;
+
+			/*
+			 * RFC 4191 Section 3.1:
+			 * "The Router Preference and Lifetime values in a ::/0
+			 * Route Information Option override the preference and
+			 * lifetime values in the Router Advertisement header."
+			 */
+			if (opts[2] == 0) {
+				if (r->lifetime == 0 && l_get_be32(opts + 4)) {
+					/*
+					 * A ::/0 route received from a
+					 * non-default router?  Should issue
+					 * a warning?
+					 */
+					break;
+				}
+
+				r->pref = preference;
+				r->lifetime = l_get_be16(opts + 4) ? 0xffff :
+					l_get_be16(opts + 6);
+				break;
+			}
+
+			/*
+			 * Don't check or warn if the route lifetime is longer
+			 * than the router lifetime because that refers to its
+			 * time as the default router.  It may be configured to
+			 * route packets for us for specific prefixes without
+			 * being a default router.
+			 */
+			i->prefix_len = opts[2];
+			i->onlink = false;
+			i->preference = preference;
+			i->valid_lifetime = l_get_be32(opts + 4);
+
+			/*
+			 * Only the initial Prefix Length bits of the prefix
+			 * are valid.  The remaining bits "MUST" be ignored
+			 * by the receiver.
+			 */
+			memcpy(i->address, net_prefix_from_ipv6(opts + 8,
+							i->prefix_len), 16);
+
+			n_routes += 1;
 			break;
 		}
 		}
 
 		opts += l;
-		opts_len -= l ;
+		opts_len -= l;
 	}
 
 	return r;
 }
 
+LIB_EXPORT char *l_icmp6_router_get_address(const struct l_icmp6_router *r)
+{
+	char buf[INET6_ADDRSTRLEN];
+
+	if (unlikely(!r))
+		return NULL;
+
+	if (!inet_ntop(AF_INET6, r->address, buf, sizeof(buf)))
+		return NULL;
+
+	return l_strdup(buf);
+}
+
 LIB_EXPORT bool l_icmp6_router_get_managed(const struct l_icmp6_router *r)
 {
 	if (unlikely(!r))
@@ -748,3 +945,20 @@ LIB_EXPORT bool l_icmp6_router_get_other(const struct l_icmp6_router *r)
 
 	return r->other;
 }
+
+LIB_EXPORT uint16_t l_icmp6_router_get_lifetime(const struct l_icmp6_router *r)
+{
+	if (unlikely(!r))
+		return false;
+
+	return r->lifetime;
+}
+
+/* Get the reception timestamp, i.e. when lifetime is counted from */
+LIB_EXPORT uint64_t l_icmp6_router_get_start_time(const struct l_icmp6_router *r)
+{
+	if (unlikely(!r))
+		return false;
+
+	return r->start_time;
+}
