@@ -39,7 +39,7 @@
 
 const char *util_ssid_to_utf8(size_t len, const uint8_t *ssid)
 {
-	static char buf[3* 32 + 1];
+	static char buf[3 * 32 + 1];
 	size_t i = 0, pos = 0;
 	const uint8_t *start = ssid, *end;
 
@@ -136,7 +136,7 @@ bool util_string_to_address(const char *str, uint8_t *out_addr)
 
 		if (str[i + 2] != ':')
 			return false;
-        }
+	}
 
 	if (!l_ascii_isxdigit(str[i]))
 		return false;
@@ -314,6 +314,7 @@ bool util_ip_prefix_tohl(const char *ip, uint8_t *prefix_out,
 struct scan_freq_set {
 	uint16_t channels_2ghz;
 	struct l_uintset *channels_5ghz;
+	struct l_uintset *channels_6ghz;
 };
 
 struct scan_freq_set *scan_freq_set_new(void)
@@ -322,13 +323,18 @@ struct scan_freq_set *scan_freq_set_new(void)
 
 	/* 802.11-2012, 8.4.2.10 hints that 200 is the largest channel number */
 	ret->channels_5ghz = l_uintset_new_from_range(1, 200);
+	ret->channels_6ghz = l_uintset_new_from_range(1, 233);
 
 	return ret;
 }
 
 void scan_freq_set_free(struct scan_freq_set *freqs)
 {
+	if (!freqs)
+		return;
+
 	l_uintset_free(freqs->channels_5ghz);
+	l_uintset_free(freqs->channels_6ghz);
 	l_free(freqs);
 }
 
@@ -347,6 +353,8 @@ bool scan_freq_set_add(struct scan_freq_set *freqs, uint32_t freq)
 		return true;
 	case BAND_FREQ_5_GHZ:
 		return l_uintset_put(freqs->channels_5ghz, channel);
+	case BAND_FREQ_6_GHZ:
+		return l_uintset_put(freqs->channels_6ghz, channel);
 	}
 
 	return false;
@@ -366,12 +374,14 @@ bool scan_freq_set_contains(const struct scan_freq_set *freqs, uint32_t freq)
 		return freqs->channels_2ghz & (1 << (channel - 1));
 	case BAND_FREQ_5_GHZ:
 		return l_uintset_contains(freqs->channels_5ghz, channel);
+	case BAND_FREQ_6_GHZ:
+		return l_uintset_contains(freqs->channels_6ghz, channel);
 	}
 
 	return false;
 }
 
-uint32_t scan_freq_set_get_bands(struct scan_freq_set *freqs)
+uint32_t scan_freq_set_get_bands(const struct scan_freq_set *freqs)
 {
 	uint32_t bands = 0;
 	uint32_t max;
@@ -384,10 +394,15 @@ uint32_t scan_freq_set_get_bands(struct scan_freq_set *freqs)
 	if (l_uintset_find_min(freqs->channels_5ghz) <= max)
 		bands |= BAND_FREQ_5_GHZ;
 
+	max = l_uintset_get_max(freqs->channels_6ghz);
+
+	if (l_uintset_find_min(freqs->channels_6ghz) <= max)
+		bands |= BAND_FREQ_6_GHZ;
+
 	return bands;
 }
 
-static void scan_channels_5ghz_add(uint32_t channel, void *user_data)
+static void scan_channels_add(uint32_t channel, void *user_data)
 {
 	struct l_uintset *to = user_data;
 
@@ -399,37 +414,41 @@ void scan_freq_set_merge(struct scan_freq_set *to,
 {
 	to->channels_2ghz |= from->channels_2ghz;
 
-	l_uintset_foreach(from->channels_5ghz, scan_channels_5ghz_add,
+	l_uintset_foreach(from->channels_5ghz, scan_channels_add,
 							to->channels_5ghz);
+	l_uintset_foreach(from->channels_6ghz, scan_channels_add,
+							to->channels_6ghz);
 }
 
 bool scan_freq_set_isempty(const struct scan_freq_set *set)
 {
-	if (set->channels_2ghz == 0 && l_uintset_isempty(set->channels_5ghz))
+	if (set->channels_2ghz == 0 && l_uintset_isempty(set->channels_5ghz) &&
+					l_uintset_isempty(set->channels_6ghz))
 		return true;
 
 	return false;
 }
 
-struct channels_5ghz_foreach_data {
+struct channels_foreach_data {
 	scan_freq_set_func_t func;
+	enum band_freq band;
 	void *user_data;
 };
 
-static void scan_channels_5ghz_frequency(uint32_t channel, void *user_data)
+static void scan_channels_foreach(uint32_t channel, void *user_data)
 {
-	const struct channels_5ghz_foreach_data *channels_5ghz_data = user_data;
+	const struct channels_foreach_data *channels_data = user_data;
 	uint32_t freq;
 
-	freq = band_channel_to_freq(channel, BAND_FREQ_5_GHZ);
+	freq = band_channel_to_freq(channel, channels_data->band);
 
-	channels_5ghz_data->func(freq, channels_5ghz_data->user_data);
+	channels_data->func(freq, channels_data->user_data);
 }
 
 void scan_freq_set_foreach(const struct scan_freq_set *freqs,
 				scan_freq_set_func_t func, void *user_data)
 {
-	struct channels_5ghz_foreach_data data = { };
+	struct channels_foreach_data data = { };
 	uint8_t channel;
 	uint32_t freq;
 
@@ -437,10 +456,14 @@ void scan_freq_set_foreach(const struct scan_freq_set *freqs,
 		return;
 
 	data.func = func;
+	data.band = BAND_FREQ_5_GHZ;
 	data.user_data = user_data;
 
-	l_uintset_foreach(freqs->channels_5ghz, scan_channels_5ghz_frequency,
-									&data);
+	l_uintset_foreach(freqs->channels_5ghz, scan_channels_foreach, &data);
+
+	data.band = BAND_FREQ_6_GHZ;
+
+	l_uintset_foreach(freqs->channels_6ghz, scan_channels_foreach, &data);
 
 	if (!freqs->channels_2ghz)
 		return;
@@ -468,9 +491,39 @@ void scan_freq_set_constrain(struct scan_freq_set *set,
 	l_uintset_free(set->channels_5ghz);
 	set->channels_5ghz = intersection;
 
+	intersection = l_uintset_intersect(constraint->channels_6ghz,
+							set->channels_6ghz);
+	if (!intersection)
+		return;
+
+	l_uintset_free(set->channels_6ghz);
+	set->channels_6ghz = intersection;
+
 	set->channels_2ghz &= constraint->channels_2ghz;
 }
 
+void scan_freq_set_subtract(struct scan_freq_set *set,
+					const struct scan_freq_set *subtract)
+{
+	struct l_uintset *sub;
+
+	sub = l_uintset_subtract(set->channels_6ghz, subtract->channels_6ghz);
+	if (L_WARN_ON(!sub))
+		return;
+
+	l_uintset_free(set->channels_6ghz);
+	set->channels_6ghz = sub;
+
+	sub = l_uintset_subtract(set->channels_5ghz, subtract->channels_5ghz);
+	if (L_WARN_ON(!sub))
+		return;
+
+	l_uintset_free(set->channels_5ghz);
+	set->channels_5ghz = sub;
+
+	set->channels_2ghz &= ~subtract->channels_2ghz;
+}
+
 static void add_foreach(uint32_t freq, void *user_data)
 {
 	uint32_t **list = user_data;
@@ -487,7 +540,8 @@ uint32_t *scan_freq_set_to_fixed_array(const struct scan_freq_set *set,
 	uint32_t *freqs;
 
 	count = __builtin_popcount(set->channels_2ghz) +
-				l_uintset_size(set->channels_5ghz);
+				l_uintset_size(set->channels_5ghz) +
+				l_uintset_size(set->channels_6ghz);
 
 	if (!count)
 		return NULL;
@@ -503,3 +557,24 @@ uint32_t *scan_freq_set_to_fixed_array(const struct scan_freq_set *set,
 
 	return freqs;
 }
+
+struct scan_freq_set *scan_freq_set_clone(const struct scan_freq_set *set,
+							uint32_t band_mask)
+{
+	struct scan_freq_set *new = l_new(struct scan_freq_set, 1);
+
+	if (band_mask & BAND_FREQ_2_4_GHZ)
+		new->channels_2ghz = set->channels_2ghz;
+
+	if (band_mask & BAND_FREQ_5_GHZ)
+		new->channels_5ghz = l_uintset_clone(set->channels_5ghz);
+	else
+		new->channels_5ghz = l_uintset_new_from_range(1, 200);
+
+	if (band_mask & BAND_FREQ_6_GHZ)
+		new->channels_6ghz = l_uintset_clone(set->channels_6ghz);
+	else
+		new->channels_6ghz = l_uintset_new_from_range(1, 233);
+
+	return new;
+}
